# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Universal logging configueration
"""

import os
import sys
import logging
import traceback
from typing import Optional, Dict

from colorama import init, Fore, Style


class LoggerSetup:  # pylint: disable=too-few-public-methods
    """ Set up the universal logging configuration """

    def __init__(self) -> None:
        if os.name == 'nt':
            init()

        formatter = ColoredFormatter('{asctime} {name} {level_color} {levelname}'
                                     '{reset}: {msg_color} {message} {reset}',
                                     style='{', datefmt='%Y-%m-%d %H:%M:%S',
                                     colors={'DEBUG': Fore.RESET + Style.BRIGHT,
                                             'INFO': Fore.RESET + Style.BRIGHT,
                                             'WARNING': Fore.YELLOW + Style.BRIGHT,
                                             'ERROR': Fore.RED + Style.BRIGHT,
                                             'CRITICAL': Fore.RED + Style.BRIGHT})

        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)

        self.logger = logging.getLogger()
        self.logger.handlers[:] = []
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        sys.excepthook = self.handle_exception

    def handle_exception(self, exc_type, exc_value, exc_traceback) -> None:  # type:ignore
        """Use Python's logging module when showing errors"""

        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        error_list = traceback.format_exception(
            exc_type, exc_value, exc_traceback)
        traceback_msg = ' '.join([str(elem) for elem in error_list[:-1]])
        error_msg = str(error_list[-1])

        # Make the error type more obvious in terminal by separating these
        self.logger.error(traceback_msg)
        self.logger.error(error_msg)


class ColoredFormatter(logging.Formatter):
    """This class formats the color of logging messages"""

    def __init__(self, *args: str, colors: Optional[Dict[str, str]] = None,
                 **kwargs: str) -> None:
        """Initialize the formatter with specified format strings"""

        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record: logging.LogRecord) -> str:
        """Format the specified record as text."""

        record.level_color = self.colors.get(record.levelname, '')
        record.reset = Style.RESET_ALL
        record.msg_color = self.colors.get(record.levelname, '')

        return super().format(record)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CLI functions for Peekingduck
"""

import os
import logging
import yaml
import click

from peekingduck.runner import Runner
from peekingduck.utils.logger import LoggerSetup
from peekingduck import __version__

LoggerSetup()
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """
    PeekingDuck is a modular computer vision inference framework.

    Developed by Computer Vision Hub at AI Singapore.
    """


def _get_cwd() -> str:
    return os.getcwd()


def create_custom_folder(custom_folder_name: str) -> None:
    """Make custom nodes folder to create custom nodes
    """
    curdir = _get_cwd()
    custom_folder_dir = os.path.join(curdir, "src", custom_folder_name)
    custom_configs_dir = os.path.join(custom_folder_dir, 'configs')

    logger.info("Creating custom nodes folder in %s", custom_folder_dir)
    os.makedirs(custom_folder_dir, exist_ok=True)
    os.makedirs(custom_configs_dir, exist_ok=True)


def create_yml() -> None:
    """Inits the declarative yaml"""
    # Default yml to be discussed
    default_yml = dict(
        nodes=[
            'input.live',
            'model.yolo',
            'draw.bbox',
            'output.screen'
        ]
    )

    with open('run_config.yml', 'w') as yml_file:
        yaml.dump(default_yml, yml_file, default_flow_style=False)


@cli.command()
@click.option('--custom_folder_name', default='custom_nodes')
def init(custom_folder_name: str) -> None:
    """Initialise a PeekingDuck project"""
    print("Welcome to PeekingDuck!")
    create_custom_folder(custom_folder_name)
    create_yml()


@cli.command()
@click.option('--config_path', default=None, type=click.Path(),
              help="List of nodes to run. None assumes \
                   run_config.yml at current working directory")
@click.option('--node_config', default="None",
              help="""Modify node configs by wrapping desired configs in a JSON string.\n
                    Example: --node_config '{"node_name": {"param_1": var_1}}' """)
def run(config_path: str, node_config: str) -> None:
    """Runs PeekingDuck"""

    curdir = _get_cwd()
    if not config_path:
        config_path = os.path.join(curdir, "run_config.yml")

    runner = Runner(config_path, node_config, "src")
    runner.run()


@cli.command()
@click.argument("type_name", required=False)
def nodes(type_name: str = None) -> None:
    """
    Lists available nodes in PeekingDuck. When no argument is given, all
    available nodes will be listed. When the node type is given as an argument
    , all available nodes in the specified node type will be listed.

    Args:
        type_name (str): input, model, dabble, draw or output
    """

    url_prefix = "https://peekingduck.readthedocs.io/en/stable/peekingduck.pipeline.nodes."
    url_postfix = ".Node.html#peekingduck.pipeline.nodes."

    if type_name is None:
        type_of_node = ["input", "model", "dabble", "draw", "output"]
    else:
        type_of_node = [type_name]

    dir_path = os.path.dirname(os.path.realpath(__file__))
    configs_dir = os.path.join(dir_path, 'configs')

    for node_type in type_of_node:
        type_dir = os.path.join(configs_dir, node_type)
        files_name = os.listdir(type_dir)

        click.secho("\nPeekingDuck has the following ", bold=True, nl=False)
        click.secho(f"{node_type} ", fg="red", bold=True, nl=False)
        click.secho("nodes:", bold=True)

        node_names = [file_name.split(".")[0] for file_name in files_name]
        max_length = len(max(node_names, key=len))
        for num, node_name in enumerate(node_names):
            url = (
                f" {url_prefix}{node_type}.{node_name}{url_postfix}"
                f"{node_type}.{node_name}.Node"
            )

            click.secho(f"{num + 1}:", nl=False)
            click.secho(f"{node_name: <{max_length + 1}}", bold=True, nl=False)
            click.secho("Info:", fg="yellow", nl=False)
            click.secho(url)

    click.secho("\n")



# Modifications copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Original copyright (c) 2019 Zihao Zhang

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Yolo image transformation
"""

import tensorflow as tf


def transform_images(x_image: tf.Tensor, size: int) -> tf.Tensor:
    """transform image to size x size

    Input:
        - x_image: input image matrix
        - size: integer size of the image

    Output:
        - x_image: transformed image matrix
    """
    x_image = tf.image.resize(x_image, (size, size))
    x_image = x_image / 255
    return x_image



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of bboxes for efficientdet
"""

import os
import logging
import json
from typing import Dict, Any, List, Tuple
import numpy as np
import tensorflow as tf
from peekingduck.utils.graph_functions import load_graph
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.utils.model_process \
    import preprocess_image, postprocess_boxes
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.model \
    import efficientdet


class Detector:
    """Detector class to handle detection of bboxes for efficientdet
    """
    GRAPH_MODE = True

    def __init__(self, config: Dict[str, Any]) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.root_dir = config['root']

        self.effdet = self._create_effdet_model()

        classes_path = os.path.join(config['root'], config['classes_file'])
        self.class_names = {value['id'] - 1: value['name']
                            for value in json.load(open(classes_path, 'r')).values()}

    def _create_effdet_model(self) -> tf.keras.Model:
        self.model_type = self.config['model_type']
        if self.GRAPH_MODE:
            graph_path = os.path.join(self.root_dir, self.config['graph_files'][self.model_type])
            model_nodes = self.config['MODEL_NODES']
            model = load_graph(
                graph_path, inputs=model_nodes['inputs'], outputs=model_nodes['outputs'])
            self.logger.info(
                'Efficientdet graph model loaded with following configs:'
                'Model type: D%s, '
                'Score Threshold: %s, ',
                self.model_type, self.config['score_threshold'])
            return model
        # For keras model
        model = efficientdet(phi=self.model_type,
                             num_classes=self.config['num_classes'],
                             score_threshold=self.config['score_threshold'])
        model_path = os.path.join(self.root_dir, self.config['graph_files'][self.model_type])
        model_path = model_path.replace('.pb', '.h5')
        model.load_weights(model_path, by_name=True)
        self.logger.info(
            'Efficientdet keras model loaded with following configs:'
            'Model type: D%s, '
            'Score Threshold: %s, ',
            self.model_type, self.config['score_threshold'])
        return model

    @staticmethod
    def preprocess(image: np.ndarray,
                   image_size: int) -> Tuple[np.ndarray, float]:
        """Preprocessing function for efficientdet

        Args:
            image (np.ndarray): image in numpy array
            image_size (int): image size as defined in efficientdet config

        Returns:
            image (np.ndarray): the preprocessed image
            scale (float): the scale the image was resized to
        """
        image, scale = preprocess_image(image, image_size=image_size)
        return image, scale

    def postprocess(self, network_output: Tuple[np.ndarray, np.ndarray, np.ndarray],
                    scale: float,
                    img_shape: List[int],
                    detect_ids: List[int]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Postprocessing of detected bboxes for efficientdet

        Args:
            network_output (list): list of boxes, scores and labels from network
            scale (float): scale the image was resized to
            img_shape (list): height of original image
            detect_ids (list): list of label ids to be detected

        Returns:
            boxes (np.ndarray): postprocessed array of detected bboxes
            scores (np.ndarray): postprocessed array of scores
            labels (np.ndarray): postprocessed array of labels
        """
        img_h, img_w = img_shape
        boxes, scores, labels = network_output
        boxes = postprocess_boxes(boxes, scale, img_h, img_w)

        indices = np.where(scores[:] > self.config['score_threshold'])[0]

        # select those detections
        boxes = boxes[indices]
        labels = labels[indices]
        scores = scores[indices]

        detect_filter = np.where(np.isin(labels, detect_ids))
        boxes = boxes[detect_filter]
        labels = labels[detect_filter]
        scores = scores[detect_filter]

        if labels.size:
            labels = np.vectorize(self.class_names.get)(labels)
        return boxes, labels, scores

    def predict_bbox_from_image(self,
                                image: np.ndarray,
                                detect_ids: List[int]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Efficientdet bbox prediction function

        Args:
            image (np.ndarray): image in numpy array
            detect_ids (list): list of label ids to be detected

        Returns:
            boxes (np.ndarray): array of detected bboxes
            scores (np.ndarray): array of scores
            labels (np.ndarray): array of labels
        """
        img_shape = image.shape[:2]

        image_size = self.config['size'][self.model_type]
        image, scale = self.preprocess(image, image_size=image_size)

        # run network
        if self.GRAPH_MODE:
            graph_input = tf.convert_to_tensor(np.expand_dims(image, axis=0), dtype=tf.float32)
            boxes, scores, labels = self.effdet(x=graph_input)
            network_output = np.squeeze(boxes.numpy()), np.squeeze(
                scores.numpy()), np.squeeze(labels.numpy())
        else:
            boxes, scores, labels = self.effdet.predict_on_batch(  # pylint:disable=no-member
                [np.expand_dims(image, axis=0)])
            network_output = np.squeeze(boxes), np.squeeze(scores), np.squeeze(labels)

        boxes, labels, scores = self.postprocess(network_output, scale, img_shape, detect_ids)

        return boxes, labels, scores
